@@ -146,7 +146,9 @@ public:
 protected:
     void zwlr_data_control_offer_v1_offer(const QString &mime_type) override
     {
-        m_receivedFormats << mime_type;
+        if (!m_receivedFormats.contains(mime_type)) {
+            m_receivedFormats << mime_type;
+        }
     }
 
     QVariant retrieveData(const QString &mimeType, QMetaType type) const override;
@@ -157,12 +159,17 @@ private:
      */
     static bool readData(int fd, QByteArray &data);
     QStringList m_receivedFormats;
+    mutable QHash<QString, QVariant> m_data;
 };
 
 QVariant DataControlOffer::retrieveData(const QString &mimeType, QMetaType type) const
 {
     Q_UNUSED(type);
 
+    auto it = m_data.constFind(mimeType);
+    if (it != m_data.constEnd())
+        return *it;
+
     QString mime;
     if (!m_receivedFormats.contains(mimeType)) {
         if (mimeType == QStringLiteral("text/plain") && m_receivedFormats.contains(utf8Text())) {
@@ -219,13 +226,31 @@ QVariant DataControlOffer::retrieveData(const QString &mimeType, QMetaType type)
             if (mimeType == applicationQtXImageLiteral()) {
                 QImage img = QImage::fromData(data, mime.mid(mime.indexOf(QLatin1Char('/')) + 1).toLatin1().toUpper().data());
                 if (!img.isNull()) {
+                    m_data.insert(mimeType, img);
                     return img;
                 }
+            } else if (data.size() > 1 && mimeType == u"text/uri-list") {
+                const auto urls = data.split('\n');
+                QVariantList list;
+                list.reserve(urls.size());
+                for (const QByteArray &s : urls) {
+#if QT_VERSION >= QT_VERSION_CHECK(6, 7, 0)
+                    if (QUrl url(QUrl::fromEncoded(QByteArrayView(s).trimmed())); url.isValid()) {
+#else
+                    if (QUrl url(QUrl::fromEncoded(QByteArrayView(s).trimmed().toByteArray())); url.isValid()) {
+#endif
+                        list.emplace_back(std::move(url));
+                    }
+                }
+                m_data.insert(mimeType, list);
+                return list;
             }
+            m_data.insert(mimeType, data);
             return data;
         }
         close(pipeFds[0]);
     }
+
     return QVariant();
 }
 
